
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// infinite.cpp*
#include "pbrt.h"
#include "light.h"
#include "texture.h"
#include "shape.h"
#include "scene.h"
#include "mipmap.h"
#include "sunsky.dat"

#define radians(x) ((x)/180.0*M_PI)

// Define Sky Structure
const struct sunsky_t
{
	double sun_theta, sun_phi;	/* Solar position */
	double sun_dir[3];			/* Solar position in Cartesian coord. */
	double turbidity;

	double V;

	// Some coefficients...
	double perez_x[5];
	double perez_y[5];
	double perez_Y[5];

	double zenith_x;
	double zenith_y;
	double zenith_Y;
};

const struct spectrum41_t
{
	int start_wavelength;
	int end_wavelength;
	int nsamples;

	float samples[41];
	float x, y, z;
	float r, g, b;
};

// InfiniteAreaLight Declarations
class InfiniteAreaLight : public Light {
public:
	// InfiniteAreaLight Public Methods
	InfiniteAreaLight(const Transform &light2world,	const Spectrum &power, int ns, const string &texmap, const Vector &dir, const float &turb, const float &thetaS, const float &phiS);
	~InfiniteAreaLight();
	Spectrum Power(const Scene *scene) const {
		Point worldCenter;
		float worldRadius;
		scene->WorldBound().BoundingSphere(&worldCenter,
		                                    &worldRadius);
		return Lbase * radianceMap->Lookup(.5f, .5f, .5f) *
			M_PI * worldRadius * worldRadius;
	}
	bool IsDeltaLight() const { return false; }
	Spectrum Le(const RayDifferential &r) const;
	Spectrum Sample_L(const Point &p, const Normal &n,
		float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const;
	Spectrum Sample_L(const Point &p, float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, float u1, float u2,
			float u3, float u4, Ray *ray, float *pdf) const;
	float Pdf(const Point &, const Normal &, const Vector &) const;
	float Pdf(const Point &, const Vector &) const;
	Spectrum Sample_L(const Point &P, Vector *w, VisibilityTester *visibility) const;
	void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v ) const;
	void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v ) const;
	
	// Realistic Atmosphere Methods
	void Initsunsky() const;
	void Rendersunsky(Vector v) const;
	double PerezFunction(const double *lam, double theta, double gamma, double lvz,
	      double sun_theta) const;
	void chromaticity_to_spectrum(double x, double y) const;
	void spectrum_to_xyz(double *x, double *y, double *z) const;
	void InfiniteAreaLight::xyz_to_rgb(double xc, double yc, double zc,
                double *r, double *g, double *b) const;

private:
	// InfiniteAreaLight Private Data
	Spectrum Lbase;
	MIPMap<Spectrum> *radianceMap;

	mutable sunsky_t sunsky;
	mutable spectrum41_t spectrum41;
	float turbidity;
};


// InfiniteAreaLight Method Definitions
InfiniteAreaLight::~InfiniteAreaLight() {
	delete radianceMap;
}
InfiniteAreaLight
	::InfiniteAreaLight(const Transform &light2world,
		                const Spectrum &L, int ns,
						const string &texmap, const Vector &dir, const float &turb, const float &thetaS, const float &phiS)
	: Light(light2world, ns) {
	radianceMap = NULL;
	if (texmap != "") {
		int width, height;
		Spectrum *texels =
			ReadImage(texmap, &width, &height);
		if (texels)
			radianceMap =
				new MIPMap<Spectrum>(width, height, texels);
		delete[] texels;
	}
	Lbase = L;
	
	sunsky.sun_theta = thetaS;
	sunsky.sun_phi = phiS;

	sunsky.sun_dir[0] = dir[0];
	sunsky.sun_dir[1] = dir[1];
	sunsky.sun_dir[2] = dir[2];
	turbidity = turb;
	
	Initsunsky();
}
//Spectrum InfiniteAreaLight::Le(const RayDifferential &r) const {
//	Vector w = r.d;
//	// Compute infinite light radiance for direction
//	Spectrum L = Lbase;
//	if (radianceMap != NULL) {
//		Vector wh = Normalize(WorldToLight(w));
//		float s = SphericalPhi(wh) * INV_TWOPI;
//		float t = SphericalTheta(wh) * INV_PI;
//		L *= radianceMap->Lookup(s, t);
//	Initsunsky();
//	}
//	return L;
//}

Spectrum InfiniteAreaLight::Le(const RayDifferential &r) const {
	Vector w = r.d;
	// Compute infinite light radiance for direction
	Vector wh = Normalize(WorldToLight(w));
	Rendersunsky(wh);
	//Spectrum L = Lbase;
	Spectrum L;
	float rgb[3];
	rgb[0] = spectrum41.r;
	rgb[1] = spectrum41.g;
	rgb[2] = spectrum41.b;
	//xyz[0] = 3.240479f * spectrum41.x + -1.537150f * spectrum41.y + -0.498535f * spectrum41.z;
	//xyz[1] = -0.969256f * spectrum41.x + 1.875991f * spectrum41.y + 0.041556f * spectrum41.z;
	//xyz[2] = 0.055648f * spectrum41.x + -0.204043f * spectrum41.y + 1.057311f * spectrum41.z;
	//L = Spectrum(spectrum41.samples);
	L = Spectrum(rgb);
	
	return L;
}

Spectrum InfiniteAreaLight::Sample_L(const Point &p,
		const Normal &n, float u1, float u2,
		Vector *wi, float *pdf,
		VisibilityTester *visibility) const {
	// Sample cosine-weighted direction on unit sphere
	float x, y, z;
	ConcentricSampleDisk(u1, u2, &x, &y);
	z = sqrtf(max(0.f, 1.f - x*x - y*y));
	if (RandomFloat() < .5) z *= -1;
	*wi = Vector(x, y, z);
	// Compute _pdf_ for cosine-weighted infinite light direction
	*pdf = fabsf(wi->z) * INV_TWOPI;
	// Transform direction to world space
	Vector v1, v2;
	CoordinateSystem(Normalize(Vector(n)), &v1, &v2);
	*wi = Vector(v1.x * wi->x + v2.x * wi->y + n.x * wi->z,
	             v1.y * wi->x + v2.y * wi->y + n.y * wi->z,
	             v1.z * wi->x + v2.z * wi->y + n.z * wi->z);
	visibility->SetRay(p, *wi);
	return Le(RayDifferential(p, *wi));
}
float InfiniteAreaLight::Pdf(const Point &, const Normal &n,
		const Vector &wi) const {
	return AbsDot(n, wi) * INV_TWOPI;
}
Spectrum InfiniteAreaLight::Sample_L(const Point &p,
		float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const {
	*wi = UniformSampleSphere(u1, u2);
	*pdf = UniformSpherePdf();
	visibility->SetRay(p, *wi);
	return Le(RayDifferential(p, *wi));
}
float InfiniteAreaLight::Pdf(const Point &, const Vector &) const {
	return 1.f / (4.f * M_PI);
}
Spectrum InfiniteAreaLight::Sample_L(const Scene *scene,
		float u1, float u2, float u3, float u4,
		Ray *ray, float *pdf) const {
	// Choose two points _p1_ and _p2_ on scene bounding sphere
	Point worldCenter;
	float worldRadius;
	scene->WorldBound().BoundingSphere(&worldCenter,
	                                    &worldRadius);
	worldRadius *= 1.01f;
	Point p1 = worldCenter + worldRadius *
		UniformSampleSphere(u1, u2);
	Point p2 = worldCenter + worldRadius *
		UniformSampleSphere(u3, u4);
	// Construct ray between _p1_ and _p2_
	ray->o = p1;
	ray->d = Normalize(p2-p1);
	// Compute _InfiniteAreaLight_ ray weight
	Vector to_center = Normalize(worldCenter - p1);
	float costheta = AbsDot(to_center,ray->d);
	*pdf =
		costheta / ((4.f * M_PI * worldRadius * worldRadius));
	return Le(RayDifferential(ray->o, -ray->d));
}
Spectrum InfiniteAreaLight::Sample_L(const Point &p,
		Vector *wi, VisibilityTester *visibility) const {
	float pdf;
	Spectrum L = Sample_L(p, RandomFloat(), RandomFloat(),
		wi, &pdf, visibility);
	if (pdf == 0.f) return Spectrum(0.f);
	return L / pdf;
}

// Calculate Sun Position Based on Input
void InfiniteAreaLight::Initsunsky() const {

	// Compute Sky Illumination
	float theta2 = sunsky.sun_theta * sunsky.sun_theta;
	float theta3 = theta2 * sunsky.sun_theta;
	
	float T = turbidity;
	float T2 = turbidity * turbidity;

	float chi = (4.0 / 9.0 - T / 120.0) * (M_PI - 2.0 * sunsky.sun_theta);
	sunsky.zenith_Y = (4.0453 * T - 4.9710) * tan(chi) - 0.2155 * T + 2.4192;
	sunsky.zenith_Y *= 1000;	/* conversion from kcd/m^2 to cd/m^2 */

	sunsky.zenith_x =
	   ( 0.00165*theta3 - 0.00374*theta2 + 0.00208*sunsky.sun_theta + 0      ) * T2
	 + (-0.02902*theta3 + 0.06377*theta2 - 0.03202*sunsky.sun_theta + 0.00394) * T
	 + ( 0.11693*theta3 - 0.21196*theta2 + 0.06052*sunsky.sun_theta + 0.25885);

	sunsky.zenith_y = 
	   ( 0.00275*theta3 - 0.00610*theta2 + 0.00316*sunsky.sun_theta + 0      ) * T2
	 + (-0.04212*theta3 + 0.08970*theta2 - 0.04153*sunsky.sun_theta + 0.00515) * T
	 + ( 0.15346*theta3 - 0.26756*theta2 + 0.06669*sunsky.sun_theta + 0.26688);

	sunsky.perez_Y[0] =  0.17872 * T  - 1.46303;
	sunsky.perez_Y[1] = -0.35540 * T  + 0.42749;
	sunsky.perez_Y[2] = -0.02266 * T  + 5.32505;
	sunsky.perez_Y[3] =  0.12064 * T  - 2.57705;
	sunsky.perez_Y[4] = -0.06696 * T  + 0.37027;
	  
	sunsky.perez_x[0] = -0.01925 * T  - 0.25922;
	sunsky.perez_x[1] = -0.06651 * T  + 0.00081;
	sunsky.perez_x[2] = -0.00041 * T  + 0.21247;
	sunsky.perez_x[3] = -0.06409 * T  - 0.89887;
	sunsky.perez_x[4] = -0.00325 * T  + 0.04517;
	  
	sunsky.perez_y[0] = -0.01669 * T  - 0.26078;
	sunsky.perez_y[1] = -0.09495 * T  + 0.00921;
	sunsky.perez_y[2] = -0.00792 * T  + 0.21023;
	sunsky.perez_y[3] = -0.04405 * T  - 1.65369;
	sunsky.perez_y[4] = -0.01092 * T  + 0.05291;
}

void InfiniteAreaLight::Rendersunsky(Vector v) const {

	float theta, phi, gamma;
	float x, y, Y;

	spectrum41.nsamples = 41;
	spectrum41.start_wavelength = 380;
	spectrum41.end_wavelength = 780;
	for (int i = 0; i < spectrum41.nsamples; i++) {
		spectrum41.samples[i] = 0.0;
	}

	/*theta = atan2(v[2],v[1]);
	phi = atan2(v[1],v[0]);

	v[0] = sin(theta) * cos(phi);
	v[2] = -sin(theta) * sin(phi);
	v[1] = cos(theta);*/
	Vector q;
	q = v;
	v[1] = v[2];
	v[2] = q[1];

	if (v[2] < 0.0) {
		// Under the horizon
		return;
	}

	// Compute the anble between viewing direction and Sun's direction
	theta = acos(v[2]);
	if (fabs(theta) < 1.0e-6) {
		phi = 0.0;
	} else {
		phi = atan2(v[1], v[0]);
	}

	float cospi;
	cospi = sin(theta) * sin(sunsky.sun_theta) * cos(sunsky.sun_phi - phi)
	      + cos(theta) * cos(sunsky.sun_theta);
	
	if (cospi > 1.0)
		gamma = 0.0;
	else if (cospi <-1.0)
		gamma = M_PI;
	else
		gamma = acos(cospi);

	x = PerezFunction(sunsky.perez_x, theta, gamma, sunsky.zenith_x, sunsky.sun_theta);
	y = PerezFunction(sunsky.perez_y, theta, gamma, sunsky.zenith_y, sunsky.sun_theta);
	Y = PerezFunction(sunsky.perez_Y, theta, gamma, sunsky.zenith_Y, sunsky.sun_theta);
	// alternative method, more direct. Manual exposure?
	float X = (x / y) * Y;
	float Z = ((1 - x - y) / y) * Y;

	float rgb[3];
	rgb[0] =  (3.240479f * X + -1.537150f * Y + -0.498535f * Z);
	rgb[1] =  (-0.969256f * X +  1.875991f * Y +  0.041556f * Z);
	rgb[2] =  (0.055648f * X + -0.204043f * Y +  1.057311f * Z);

	// scaling from xyY -> XYZ -> RGB -> HSV -> RGB to get the best luminance scaling (in HSV)
	float expFactor = 3.0f;
	float hsv[3];
	RGBtoHSV(rgb[0], rgb[1], rgb[2], &hsv[0], &hsv[1], &hsv[2]);
	hsv[2] = 1.0f - exp(-expFactor * hsv[2]);
	HSVtoRGB(&rgb[0], &rgb[1], &rgb[2], hsv[0], hsv[1], hsv[2]);	
	
	spectrum41.r = rgb[0];
	spectrum41.g = rgb[1];
	spectrum41.b = rgb[2];


	// Old method: convert to spectrum first. Not necessary for skylight
	/*chromaticity_to_spectrum(x, y);
	double lx, ly, lz;
	spectrum_to_xyz(&lx, &ly, &lz);

	if (fabs(ly) < 1.0e-48) {
		ly = 1.0;
	}

	for (int i = 0; i < spectrum41.nsamples; i++) {
		spectrum41.samples[i] = Y * spectrum41.samples[i] / ly;
	}

	double sky_x, sky_y, sky_z, sky_r, sky_g, sky_b;
	spectrum_to_xyz(&sky_x, &sky_y, &sky_z);
	xyz_to_rgb(sky_x, sky_y, sky_z, &sky_r, &sky_g, &sky_b);
	spectrum41.x = sky_x;
	spectrum41.y = sky_y;
	spectrum41.z = sky_z;
	spectrum41.r = sky_r;
	spectrum41.g = sky_g;
	spectrum41.b = sky_b;	*/
}

double InfiniteAreaLight::PerezFunction(const double *lam, double theta, double gamma, double lvz,
	      double sun_theta) const {
	double den;
	double num;
	
	den = ((1.0 + lam[0] * exp(lam[1])) *
	       (1.0 + lam[2] * exp(lam[3] * sun_theta) + lam[4] * cos(sun_theta) * cos(sun_theta)));

	num = ((1.0 + lam[0] * exp(lam[1]/cos(theta))) *
	       (1.0 + lam[2] * exp(lam[3] * gamma) + lam[4] * cos(gamma) * cos(gamma)));

	return lvz * num / den;
}

void InfiniteAreaLight::chromaticity_to_spectrum(double x, double y) const {
	int i;
	double M1, M2;

	M1 = (-1.3515 - 1.7703 * x +  5.9114 * y)
           / (0.0241 + 0.2562 * x - 0.7341 * y);
	M2 = (0.03 - 31.4424 * x + 30.0717 * y)
           / (0.0241 + 0.2562 * x - 0.7341 * y);

	// Assume that spectrum range is 380nm - 780nm by 10nm, 41 samples
	for (i = 0; i < spectrum41.nsamples; i++) {
		spectrum41.samples[i] = S0Amplitudes[i] + M1 * S1Amplitudes[i] + M2 * S2Amplitudes[i];
	} 
}

void InfiniteAreaLight::spectrum_to_xyz(double *x, double *y, double *z) const {
    int i;
    double lambda, X = 0, Y = 0, Z = 0, XYZ;

    static double cie_colour_match[81][3] = {
        {0.0014,0.0000,0.0065}, {0.0022,0.0001,0.0105}, {0.0042,0.0001,0.0201},
        {0.0076,0.0002,0.0362}, {0.0143,0.0004,0.0679}, {0.0232,0.0006,0.1102},
        {0.0435,0.0012,0.2074}, {0.0776,0.0022,0.3713}, {0.1344,0.0040,0.6456},
        {0.2148,0.0073,1.0391}, {0.2839,0.0116,1.3856}, {0.3285,0.0168,1.6230},
        {0.3483,0.0230,1.7471}, {0.3481,0.0298,1.7826}, {0.3362,0.0380,1.7721},
        {0.3187,0.0480,1.7441}, {0.2908,0.0600,1.6692}, {0.2511,0.0739,1.5281},
        {0.1954,0.0910,1.2876}, {0.1421,0.1126,1.0419}, {0.0956,0.1390,0.8130},
        {0.0580,0.1693,0.6162}, {0.0320,0.2080,0.4652}, {0.0147,0.2586,0.3533},
        {0.0049,0.3230,0.2720}, {0.0024,0.4073,0.2123}, {0.0093,0.5030,0.1582},
        {0.0291,0.6082,0.1117}, {0.0633,0.7100,0.0782}, {0.1096,0.7932,0.0573},
        {0.1655,0.8620,0.0422}, {0.2257,0.9149,0.0298}, {0.2904,0.9540,0.0203},
        {0.3597,0.9803,0.0134}, {0.4334,0.9950,0.0087}, {0.5121,1.0000,0.0057},
        {0.5945,0.9950,0.0039}, {0.6784,0.9786,0.0027}, {0.7621,0.9520,0.0021},
        {0.8425,0.9154,0.0018}, {0.9163,0.8700,0.0017}, {0.9786,0.8163,0.0014},
        {1.0263,0.7570,0.0011}, {1.0567,0.6949,0.0010}, {1.0622,0.6310,0.0008},
        {1.0456,0.5668,0.0006}, {1.0026,0.5030,0.0003}, {0.9384,0.4412,0.0002},
        {0.8544,0.3810,0.0002}, {0.7514,0.3210,0.0001}, {0.6424,0.2650,0.0000},
        {0.5419,0.2170,0.0000}, {0.4479,0.1750,0.0000}, {0.3608,0.1382,0.0000},
        {0.2835,0.1070,0.0000}, {0.2187,0.0816,0.0000}, {0.1649,0.0610,0.0000},
        {0.1212,0.0446,0.0000}, {0.0874,0.0320,0.0000}, {0.0636,0.0232,0.0000},
        {0.0468,0.0170,0.0000}, {0.0329,0.0119,0.0000}, {0.0227,0.0082,0.0000},
        {0.0158,0.0057,0.0000}, {0.0114,0.0041,0.0000}, {0.0081,0.0029,0.0000},
        {0.0058,0.0021,0.0000}, {0.0041,0.0015,0.0000}, {0.0029,0.0010,0.0000},
        {0.0020,0.0007,0.0000}, {0.0014,0.0005,0.0000}, {0.0010,0.0004,0.0000},
        {0.0007,0.0002,0.0000}, {0.0005,0.0002,0.0000}, {0.0003,0.0001,0.0000},
        {0.0002,0.0001,0.0000}, {0.0002,0.0001,0.0000}, {0.0001,0.0000,0.0000},
        {0.0001,0.0000,0.0000}, {0.0001,0.0000,0.0000}, {0.0000,0.0000,0.0000}
    };

    for (i = 0, lambda = 380; lambda < 780.1; i+=2, lambda += 10) {

        X += spectrum41.samples[i / 2] * cie_colour_match[i][0];
        Y += spectrum41.samples[i / 2] * cie_colour_match[i][1];
        Z += spectrum41.samples[i / 2] * cie_colour_match[i][2];

    }

	XYZ = (X + Y + Z);
	XYZ = 1.0;
    if (fabs(XYZ) > 1.0e-48) { 
        *x = X / XYZ;
        *y = Y / XYZ;
        *z = Z / XYZ;
    } else {
        *x = X;
        *y = Y;
        *z = Z;
	}
}

void InfiniteAreaLight::xyz_to_rgb(double xc, double yc, double zc,
                double *r, double *g, double *b) const {
    double xr, yr, zr, xg, yg, zg, xb, yb, zb;
    double xw, yw, zw;
    double rx, ry, rz, gx, gy, gz, bx, by, bz;
    double rw, gw, bw;

	// CIE color system definitions
	double xRed = 0.7355;
	double yRed = 0.2645;
	double xGreen = 0.2658;
	double yGreen = 0.7243;
	double xBlue = 0.1669;
	double yBlue = 0.0085;
	double xWhite = 0.33333333;
	double yWhite = 0.33333333;
	double gamma = 0.0;

    xr = xRed;    yr = yRed;    zr = 1 - (xr + yr);
    xg = xGreen;  yg = yGreen;  zg = 1 - (xg + yg);
    xb = xBlue;   yb = yBlue;   zb = 1 - (xb + yb);

    xw = xWhite;  yw = yWhite;  zw = 1 - (xw + yw);

    /* xyz -> rgb matrix, before scaling to white. */
    
    rx = (yg * zb) - (yb * zg);  ry = (xb * zg) - (xg * zb);  rz = (xg * yb) - (xb * yg);
    gx = (yb * zr) - (yr * zb);  gy = (xr * zb) - (xb * zr);  gz = (xb * yr) - (xr * yb);
    bx = (yr * zg) - (yg * zr);  by = (xg * zr) - (xr * zg);  bz = (xr * yg) - (xg * yr);

    /* White scaling factors.
       Dividing by yw scales the white luminance to unity, as conventional. */
       
    if (fabs(yw) > 1.0e-48) {
      rw = ((rx * xw) + (ry * yw) + (rz * zw)) / yw;
      gw = ((gx * xw) + (gy * yw) + (gz * zw)) / yw;
      bw = ((bx * xw) + (by * yw) + (bz * zw)) / yw;
    } else {
      rw = 1.0;
      gw = 1.0;
      bw = 1.0;
    }

    /* xyz -> rgb matrix, correctly scaled to white. */
    
    rx = rx / rw;  ry = ry / rw;  rz = rz / rw;
    gx = gx / gw;  gy = gy / gw;  gz = gz / gw;
    bx = bx / bw;  by = by / bw;  bz = bz / bw;

    /* rgb of the desired point */
    
    *r = (rx * xc) + (ry * yc) + (rz * zc);
    *g = (gx * xc) + (gy * yc) + (gz * zc);
    *b = (bx * xc) + (by * yc) + (bz * zc);
}

void InfiniteAreaLight::RGBtoHSV( float r, float g, float b, float *h, float *s, float *v ) const {
	float min, max, delta;

	if ( r < g ) {
		if ( r < b ) min = r; else min = b;
	} else {
		if ( g < b ) min = g; else min = b;
	}

	if ( r > g ) {
		if ( r > b ) max = r; else max = b;
	} else {
		if ( g > b ) max = g; else max = b;
	}

	*v = max;				// v

	delta = max - min;

	if( max != 0.f )
		*s = delta / max;		// s
	else {
		// r = g = b = 0		// s = 0, v is undefined
		*s = 0.f;
		*h = -1.f;
		return;
	}

	if( r == max )
		*h = ( g - b ) / delta;		// between yellow & magenta
	else if( g == max )
		*h = 2.f + ( b - r ) / delta;	// between cyan & yellow
	else
		*h = 4.f + ( r - g ) / delta;	// between magenta & cyan

	*h *= 60.f;				// degrees
	if( *h < 0 )
		*h += 360.f;

}

void InfiniteAreaLight::HSVtoRGB( float *r, float *g, float *b, float h, float s, float v ) const {
	int i;
	float f, p, q, t;

	if( s == 0.f ) {
		// achromatic (grey)
		*r = *g = *b = v;
		return;
	}

	h /= 60.f;			// sector 0 to 5
	i = floor( h );
	f = h - i;			// factorial part of h
	p = v * ( 1.f - s );
	q = v * ( 1.f - s * f );
	t = v * ( 1.f - s * ( 1 - f ) );

	switch( i ) {
		case 0:
			*r = v;
			*g = t;
			*b = p;
			break;
		case 1:
			*r = q;
			*g = v;
			*b = p;
			break;
		case 2:
			*r = p;
			*g = v;
			*b = t;
			break;
		case 3:
			*r = p;
			*g = q;
			*b = v;
			break;
		case 4:
			*r = t;
			*g = p;
			*b = v;
			break;
		default:		// case 5:
			*r = v;
			*g = p;
			*b = q;
			break;
	}

}

extern "C" DLLEXPORT Light *CreateLight(const Transform &light2world,
		const ParamSet &paramSet) {
	//Spectrum L = paramSet.FindOneSpectrum("L", Spectrum(1.0));
	Spectrum L;
	L.Black();
	string texmap = paramSet.FindOneString("mapname", "");
	int nSamples = paramSet.FindOneInt("nsamples", 1);

	float timeOfDay = paramSet.FindOneFloat("timeOfDay", 12.0f);
	float julianDay = paramSet.FindOneFloat("julianDay", 78.0f);
	float standardMeridian = paramSet.FindOneFloat("standardMeridian", 9.0f);
	float longitude = paramSet.FindOneFloat("longitude", 139.44f);
	float latitude = paramSet.FindOneFloat("latitude", 35.53f);
	float turb = paramSet.FindOneFloat("turbidity", 4.0f); 
	
	// Compute sun position
	
	float phiS, thetaS;
	float theta2, theta3, T, T2, chi;
	
	standardMeridian = radians(standardMeridian * 15);
	float solarTime = timeOfDay +
	(0.170*sin(4*M_PI*(julianDay - 80)/373) - 0.129*sin(2*M_PI*(julianDay - 8)/355)) +
	12.0 * (standardMeridian - radians(longitude))/M_PI;

    float solarDeclination = (0.4093*sin(2*M_PI*(julianDay - 81)/368));

    float solarAltitude= asin(sin(radians(latitude)) * sin(solarDeclination) -
			       cos(radians(latitude)) * cos(solarDeclination) * cos(M_PI*solarTime/12));

    float opp, adj;
    opp = -cos(solarDeclination) * sin(M_PI*solarTime/12);
    adj = -(cos(radians(latitude)) * sin(solarDeclination) +
	    sin(radians(latitude)) * cos(solarDeclination) *  cos(M_PI*solarTime/12));
    float solarAzimuth=atan2(opp,adj);

    phiS = -solarAzimuth;
    thetaS = M_PI / 2.0 - solarAltitude;

	Vector dir;
	dir.x = cos(phiS) * sin(thetaS);
	dir.y = sin(phiS) * sin(thetaS);
	dir.z = cos(thetaS);

	return new InfiniteAreaLight(light2world, L, nSamples, texmap, dir, turb, thetaS, phiS);
}
